//==============================
//	qe3.h
//==============================

#ifndef __QE3_H__
#define __QE3_H__

// disable data conversion warnings for gl
#pragma warning(disable : 4244)     // MIPS
#pragma warning(disable : 4136)     // X86
#pragma warning(disable : 4051)     // ALPHA
 
#include <windows.h>
#include <commctrl.h>
#include "afxres.h"
#include "resource.h"

#include <gl/gl.h>
#include <gl/glu.h>
#include <gl/glaux.h>
#include <math.h>
#include <stdlib.h>
#include <malloc.h>	// sikk - Undo/Redo

#include "cmdlib.h"
#include "lbmlib.h"
#include "mathlib.h"
#include "parse.h"

#include "qedefs.h"
#include "qfiles.h"

#include "textures.h"
#include "brush.h"
#include "clip.h"
#include "csg.h"
#include "drag.h"
#include "entity.h"
#include "map.h"
#include "points.h"
#include "select.h"
#include "vertsel.h"
#include "winding.h"

#include "camera.h"
#include "xy.h"
#include "z.h"
#include "mru.h"
#include "undo.h"

//========================================================================

typedef struct
{
	int		nSize;				// structure size
	int		nTexMenu;			// nearest, linear, etc
//	int		nVTexMenu;			// sikk - unused
	int		nExclude;
	bool	bShow_Axis,			// sikk - Show Axis
			bShow_Blocks,
			bShow_CameraGrid,	// sikk - Show Camera Grid
			bShow_Coordinates,
			bShow_LightRadius,	// sikk - Show Light Radius
			bShow_MapBoundry,	// sikk - Show Map Boundry Box
			bShow_Names,
			bShow_SizeInfo,
			bShow_Viewname,		// sikk - Show View Name
			bShow_Workzone,
			bNoClamp,
			bScaleLockX,		// sikk - Brush Scaling Axis Lock
			bScaleLockY, 
			bScaleLockZ,
			bCubicClip;			// sikk - Cubic Clipping
	int		nCubicScale;		// sikk - Cubic Clipping
	int		nCameraSpeed;		// sikk - Camera Speed Trackbar 
	vec3_t	v3Colors[COLOR_LAST];

// sikk---> Preferences Dialog
	char	szGamePath[_MAX_PATH],
			szGameName[_MAX_FNAME],
			szPrefabPath[_MAX_PATH],
			szLastProject[_MAX_PATH],
			szLastMap[_MAX_PATH],
			szModName[_MAX_FNAME],
			szHeapsize[8],
			szSkill[1];
	int		nAutosave,
			nHeapsize,
			nMapSize,
			nUndoLevels;
	float	fGamma;				// gamma for textures
	bool	bAutosave,
			bLoadLastProject,
			bLoadLastMap,
			bLogConsole,
			bNoStipple,
			bRadiantLights,
			bVertexSplitsFace,
			bModName,
			bHeapsize,
			bSkill,
			bDeathmatch,
			bDeveloper,
			bRSpeeds,
			bPointfile,
			bTestAfterBSP;
// <---sikk
} savedinfo_t;

/*
** most of the QE globals are stored in this structure
*/
typedef struct
{
	bool		d_bShowGrid;
	int			d_nGridSize;

	vec3_t		d_v3WorkMin, 	// defines the boundaries of the current work area is used to guess
				d_v3WorkMax;	// brushes and drop points third coordinate when creating from 2D view

	HGLRC		d_hglrcBase;
	HDC			d_hdcBase;
	HINSTANCE	d_hInstance;
	HWND		d_hInstanceColor;		//eerie
	HWND		d_hwndMain;
	HWND		d_hwndCamera;
	HWND		d_hwndConsole;
	HWND		d_hwndEntity;
	HWND		d_hwndTexture;
	HWND		d_hwndXY;
	HWND		d_hwndZ;
	HWND		d_hwndStatus;
	HWND		d_hwndToolbar1;
	HWND		d_hwndToolbar2;
	HWND		d_hwndToolbar3;
	HWND		d_hwndToolbar4;
	HWND		d_hwndToolbar5;
	HWND		d_hwndToolbar6;
	HWND		d_hwndToolbar7;
	HWND		d_hwndToolbar8;
	HWND		d_hwndRebar;			// sikk - Rebar 
	HWND		d_hwndSplash;			// sikk - Splash screen

	entity_t   *d_entityProject;
	int			d_nNumEntities;

	int			d_nNumPoints;
	int			d_nNumEdges;
	int			d_nNumMovePoints;
	vec3_t		d_v3Points[MAX_POINTS];
	pedge_t		d_pEdges[MAX_EDGES];
	float      *d_fMovePoints[1024];

	qtexture_t *d_qtextures;
	texturewin_t d_texturewin;

	int			d_nPointfileDisplayList;

	camera_t	d_camera;				// sikk - moved camera object here
	xy_t        d_xyz;					// it's now for all view
	z_t         d_z;

	int         d_nViewType;			// XY = x0, y1; XZ = x0, y2; YZ = x1, y2.
	int         d_nInspectorMode;		// W_TEXTURE, W_ENTITY, or W_CONSOLE

	LPMRUMENU   d_lpMruMenu;

	savedinfo_t d_savedinfo;

	int         d_nWorkCount;

	// connect entities uses the last two brushes selected
	int			d_nSelectCount;
	brush_t	   *d_pbrSelectOrder[2];
	vec3_t      d_v3SelectTranslate;    // for dragging w/o making new display lists
	select_t    d_selSelectMode;

	int		    d_nFontList;
	int         d_nParsedBrushes;

	bool		d_bTextureLock;

	bool	    d_bClipMode;
	bool	    d_bClipSwitch;
	brush_t     d_brFrontSplits;
	brush_t     d_brBackSplits;
	brush_t    *d_pbrSplitList;

	// handle to the console log file
	// we use low level I/O to get rid of buffering and have everything on file if we crash
	int         d_nLogFile;

	bool		d_bResetRegistry;		// sikk - this is used for Preferences 'Reset Registry' command
} qeglobals_t;

//========================================================================

/*
** extern declarations
*/
extern qeglobals_t g_qeglobals;

extern int		g_nUpdateBits;
extern int		g_nScreenWidth;
extern int		g_nScreenHeight;
extern char    *g_szBSP_Commands[256];
extern HANDLE	g_hBSP_Process;

/*
** global declarations
*/
int	g_nNumBrushes, g_nNumEntities, g_nNumTextures;

//========================================================================

/*
** QE function declarations
*/
void	QE_CheckAutoSave ();
void	QE_CheckOpenGLForErrors (void);
void	QE_ConvertDOSToUnixName (char *dst, const char *src);
void	QE_CountBrushesAndUpdateStatusBar ();
void	QE_ExpandBspString (char *bspaction, char *out, char *mapname);
void	QE_Init ();
bool	QE_KeyDown (int key);
bool	QE_LoadProject (char *projectfile);
bool	QE_SingleBrush ();
void	QE_UpdateCommandUI ();
char   *QE_ExpandRelativePath (char *p);

/*
** QE Win32 function declarations
*/
int		QEW_SetupPixelFormat (HDC hDC, bool zbuffer);
void	QEW_StopGL (HWND hWnd, HGLRC hGLRC, HDC hDC);

void   *qmalloc (int size);
char   *CopyString (char *s);

//
// system functions
//
void    Sys_UpdateBrushStatusBar ();
void	Sys_UpdateGridStatusBar ();
void    Sys_UpdateWindows (int bits);
void    Sys_Beep ();
void    Sys_ClearPrintf ();
void    Sys_Printf (char *text, ...);
double	Sys_DoubleTime ();
void    Sys_GetCursorPos (int *x, int *y);
void    Sys_SetCursorPos (int x, int y);
void    Sys_SetTitle (char *text);
void    Sys_BeginWait ();
void    Sys_EndWait ();
void    Sys_Status (const char *psz, int part);
void	Sys_LogFile ();

//
// win_qe3.c
//
void	FillBSPMenu ();
char   *TranslateString (char *buf);
void	ProjectDialog ();
void	NewProjectDialog ();	// sikk - New Project Dialog
void	OpenDialog ();
void	SaveAsDialog ();
bool	ConfirmModified ();
void	ImportDialog (bool bCheck);	// sikk - Import Dialog for map/prefab
void	ExportDialog (bool bCheck);	// sikk - Export Dialog for map/prefab

//
// textures.c
//
HWND WTex_Create (HINSTANCE hInstance);
LONG WINAPI WTex_WndProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

//
// win_ent.c
//
void WEnt_Create (HINSTANCE hInstance);

BOOL CALLBACK FieldWndProc (HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK EntityListWndProc (HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK EntityWndProc (HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);

LRESULT (CALLBACK* OldFieldWindowProc) (HWND, UINT, WPARAM, LPARAM);
LRESULT (CALLBACK* OldEntityListWindowProc) (HWND, UINT, WPARAM, LPARAM);

void CreateEntityWindow (HINSTANCE hInstance);
bool UpdateEntitySel (eclass_t *pec);	
bool UpdateSel (int nIndex, eclass_t *pec);
void CreateEntity ();
void FillClassList ();
void AddProp ();
void DelProp ();
void EditProp ();
void GetEntityControls (HWND h);
void GetSpawnFlags ();
void SetSpawnFlags ();
void SetKeyValuePairs ();
void SetInspectorMode (int nType);
void SizeEntityDlg (int nWidth, int nHeight);
void MOVE (HWND e, int x, int y, int w, int h);

//
// win_cam.c
//
void WCam_Create (HINSTANCE hInstance);
LONG WINAPI WCam_WndProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

//
// win_xy.c
//
void WXY_Create (HINSTANCE hInstance);
LONG WINAPI WXY_WndProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
//static void WXY_InitPixelFormat (PIXELFORMATDESCRIPTOR *pPFD);	// sikk - unused
void WXY_Print ();
int  GetSelectionInfo ();	// sikk - Contex Menu
void DoXYPopupMenu (int x, int y);	// sikk - Contex Menu

//
// win_z.c
//
void WZ_Create (HINSTANCE hInstance);
LONG WINAPI WZ_WndProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

//
// win_main.c
//
void WMain_Create (HINSTANCE hInstance);
LONG WINAPI CommandHandler (HWND hWnd, WPARAM wParam, LPARAM lParam); // sikk - Declaration for Popup menu
bool DoColor (int iIndex);
void DoTestMap ();	// sikk - Test Map

static HWND CreateStatusBar (HWND hWnd);
static HWND CreateReBar (HWND hWnd, HINSTANCE hInst);	// sikk - Rebar
static HWND CreateToolBar (HWND hWnd, HINSTANCE hInst, int nIndex, int nPos, int nButtons, int nSize); // sikk - Toolbars
static HWND CreateTrackBar (HWND hWnd, HINSTANCE hInst, int nIndex);	// sikk - Camera Speed Trackbar

extern BOOL SaveWindowState (HWND hWnd, const char *pszName);
extern BOOL LoadWindowState (HWND hWnd, const char *pszName);
extern BOOL SaveRegistryInfo (const char *pszName, void *pvBuf, long lSize);
extern BOOL LoadRegistryInfo (const char *pszName, void *pvBuf, long *plSize);

//
// win_dlg.c
//
void DoFindBrush ();
void DoRotate ();
void DoSides (int nType);	// sikk - Brush Primitives (previously took no arguments)
void DoKeylist ();
void DoAbout ();
void DoSurface ();
void DoFindTexture ();
void DoNewProject();	// sikk - New Project Dialog
void DoProject (bool bFirst);	// sikk - Project Settings Dialog
void DoCreateEntity (bool bPointbased, bool bBrushbased, bool bSel, vec3_t origin);	// sikk - Create Entity Dialog
void DoMapInfo ();	// sikk - Map Info Dialog
void DoEntityInfo ();	// sikk - Entity Info Dialog
void DoPreferences ();	// sikk - Preferences Dialog
void DoScale ();	// sikk - Brush Scaling Dialog
void DoCamSpeed ();

//
// win_proj.c
//
BOOL CALLBACK ProjectSettingsDlgProc (HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK SelectDirDlgProc (HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);

void GetProjectDirectory (HWND hwndDlg);
void GetRemoteBasePath (HWND hwndDlg);
void GetMapsDirectory (HWND hwndDlg);
void GetAutosaveMap (HWND hwndDlg);
void GetEntityFiles (HWND hwndDlg);
void GetTextureDirectory (HWND hwndDlg);
void GetToolsDirectory (HWND hwndDlg);

int  GetNextFreeBspIndex ();
int  GetBspIndex (char *text);
void NewBspCommand (HWND hwndDlg);
void AcceptBspCommand (HWND hwndDlg);
void DeleteBspCommand (HWND hwndDlg);

void SaveSettings (HWND hwndDlg);
bool SelectDir (HWND h);

//
// win_snap.c
//
bool FindClosestBottom (HWND h, LPRECT rect, LPRECT parentrect);
bool FindClosestTop (HWND h, LPRECT rect, LPRECT parentrect);
bool FindClosestLeft (HWND h, LPRECT rect, LPRECT parentrect, int widthlimit);
bool FindClosestRight (HWND h, LPRECT rect, LPRECT parentrect, int widthlimit);
bool FindClosestHorizontal (HWND h, LPRECT rect, LPRECT parentrect);
bool FindClosestVertical (HWND h, LPRECT rect, LPRECT parentrect);
bool TryDocking (HWND h, long side, LPRECT rect, int widthlimit); // sikk - Window Management

#endif
